@@ -19,10 +19,12 @@ with this program.  If not, see <http://www.gnu.org/licenses/>.
 
 
 #include <glib.h>
+#include <gio/gio.h>
 
 static GList * tasks = NULL;
 static gboolean global_success = TRUE;
 static GMainLoop * global_mainloop;
+static gint max_wait = 30;
 
 typedef enum {
 	TASK_RETURN_NORMAL = 0,
@@ -42,16 +44,21 @@ typedef struct {
 	gboolean text_die;
 	guint idle;
 	guint io_watch;
+	gchar * wait_for;
 } task_t;
 
 static void check_task_cleanup (task_t * task, gboolean force);
+static void start_task (gpointer data, gpointer userdata);
+static void bustle_watcher (GPid pid, gint status, gpointer data);
 
+static gchar * bustle_cmd = NULL;
 static gchar * bustle_datafile = NULL;
+static guint bustle_watch = 0;
 static GIOChannel * bustle_stdout = NULL;
 static GIOChannel * bustle_stderr = NULL;
 static GIOChannel * bustle_file = NULL;
 static GPid bustle_pid = 0;
-static GList * bustle_watches = NULL;
+static gboolean any_waitfors = FALSE;
 
 #define BUSTLE_ERROR_DEFAULT  "Bustle"
 static gchar * bustle_error = BUSTLE_ERROR_DEFAULT;
@@ -122,29 +129,27 @@ start_bustling (void)
 	bustle_file = g_io_channel_new_file(bustle_datafile, "w", &error);
 
 	if (error != NULL) {
-		g_warning("Unable to open bustle file '%s': %s", bustle_datafile, error->message);
+		g_critical("Unable to open bustle file '%s': %s", bustle_datafile, error->message);
 		g_error_free(error);
 		g_free(bustle_datafile);
 		bustle_datafile = NULL;
+		global_success = FALSE;
+		g_main_loop_quit(global_mainloop);
 		return;
 	}
 
 	gint bustle_stdout_num;
 	gint bustle_stderr_num;
 	
-	gchar ** bustle_monitor = g_new0(gchar *, g_list_length(bustle_watches) + 3);
-	bustle_monitor[0] = "bustle-dbus-monitor";
+	gchar ** bustle_monitor = g_new0(gchar *, 3);
+	bustle_monitor[0] = (gchar *)bustle_cmd;
 	bustle_monitor[1] = "--session";
-	int i;
-	for (i = 0; i < g_list_length(bustle_watches); i++) {
-		bustle_monitor[i + 2] = (gchar *)g_list_nth(bustle_watches, i)->data;
-	}
 
 	g_spawn_async_with_pipes(g_get_current_dir(),
 	                         bustle_monitor, /* argv */
 	                         NULL, /* envp */
 	                         /* G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL, */ /* flags */
-	                         G_SPAWN_SEARCH_PATH, /* flags */
+	                         G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD, /* flags */
 	                         NULL, /* child setup func */
 	                         NULL, /* child setup data */
 	                         &bustle_pid, /* PID */
@@ -154,11 +159,16 @@ start_bustling (void)
 	                         &error); /* error */
 
 	if (error != NULL) {
-		g_warning("Unable to start bustling data: %s", error->message);
+		g_critical("Unable to start bustling data: %s", error->message);
+		g_error_free(error);
+		bustle_pid = 0;
+		global_success = FALSE;
+		g_main_loop_quit(global_mainloop);
 		return;
 	}
 
 	g_debug("Starting bustle monitor.  PID: %d", bustle_pid);
+	bustle_watch = g_child_watch_add(bustle_pid, bustle_watcher, NULL);
 
 	bustle_stdout = g_io_channel_unix_new(bustle_stdout_num);
 	g_io_add_watch(bustle_stdout,
@@ -182,6 +192,14 @@ stop_bustling (void)
 		return;
 	}
 
+	if (bustle_watch != 0) {
+		g_source_remove(bustle_watch);
+	}
+
+	if (bustle_pid == 0) {
+		return;
+	}
+
 	gchar * killline = g_strdup_printf("kill -INT %d", bustle_pid);
 	g_spawn_command_line_sync(killline, NULL, NULL, NULL, NULL);
 	g_free(killline);
@@ -263,6 +281,7 @@ check_task_cleanup (task_t * task, gboolean force)
 	}
 
 	tasks = g_list_remove(tasks, task);
+	g_free(task->wait_for);
 	g_free(task->executable);
 	g_free(task->name);
 	g_free(task);
@@ -274,6 +293,36 @@ check_task_cleanup (task_t * task, gboolean force)
 	return;
 }
 
+static void
+dbus_watcher (GPid pid, gint status, gpointer data)
+{
+	g_critical("DBus Daemon exited abruptly!");
+
+	global_success = FALSE;
+	g_main_loop_quit(global_mainloop);
+
+	if (pid != 0) {
+		g_spawn_close_pid(pid);
+	}
+
+	return;
+}
+
+static void
+bustle_watcher (GPid pid, gint status, gpointer data)
+{
+	g_critical("Bustle Monitor exited abruptly!");
+
+	global_success = FALSE;
+	g_main_loop_quit(global_mainloop);
+
+	if (pid != 0) {
+		g_spawn_close_pid(pid);
+	}
+
+	return;
+}
+
 static void
 proc_watcher (GPid pid, gint status, gpointer data)
 {
@@ -327,12 +376,30 @@ proc_writes (GIOChannel * channel, GIOCondition condition, gpointer data)
 	return TRUE;
 }
 
-void
+static void
+wait_for_found (GDBusConnection * connection, const gchar * name, const gchar * name_owner, gpointer user_data)
+{
+	return start_task(user_data, GINT_TO_POINTER(TRUE));
+}
+
+static void
 start_task (gpointer data, gpointer userdata)
 {
+	gboolean ignore_waitfor = GPOINTER_TO_INT(userdata);
 	GError * error = NULL;
 	task_t * task = (task_t *)data;
 
+	if (!ignore_waitfor && task->wait_for != NULL) {
+		g_bus_watch_name(G_BUS_TYPE_SESSION,
+		                 task->wait_for,
+		                 G_BUS_NAME_WATCHER_FLAGS_NONE,
+		                 wait_for_found,
+		                 NULL,
+		                 task,
+		                 NULL);
+		return;
+	}
+
 	gchar ** argv;
 	argv = g_new0(gchar *, g_list_length(task->parameters) + 2);
 
@@ -380,7 +447,7 @@ gboolean
 dbus_writes (GIOChannel * channel, GIOCondition condition, gpointer data)
 {
 	if (condition & G_IO_ERR) {
-		g_error("DBus writing failure!");
+		g_critical("DBus writing failure!");
 		return FALSE;
 	}
 
@@ -403,9 +470,14 @@ dbus_writes (GIOChannel * channel, GIOCondition condition, gpointer data)
 		if (tasks != NULL) {
 			start_bustling();
 
-			g_list_foreach(tasks, start_task, NULL);
+			/* If we're still in a place where we can succeed, then
+			   we should continue.  Otherwise fail. */
+			if (global_success) {
+				g_list_foreach(tasks, start_task, GINT_TO_POINTER(FALSE));
+			}
 		} else {
 			g_print("No tasks!\n");
+			global_success = FALSE;
 			g_main_loop_quit(global_mainloop);
 		}
 	}
@@ -425,6 +497,7 @@ option_task (const gchar * arg, const gchar * value, gpointer data, GError ** er
 	task->task_die = FALSE;
 	task->text_die = FALSE;
 	tasks = g_list_prepend(tasks, task);
+	task->wait_for = NULL;
 	return TRUE;
 }
 
@@ -497,9 +570,21 @@ option_param (const gchar * arg, const gchar * value, gpointer data, GError ** e
 }
 
 static gboolean
-bustle_watch (const gchar * arg, const gchar * value, gpointer data, GError ** error)
+option_wait (const gchar * arg, const gchar * value, gpointer data, GError ** error)
 {
-	bustle_watches = g_list_append(bustle_watches, g_strdup(value));
+	if (tasks == NULL) {
+		g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, "No task to add a wait on %s for.", value);
+		return FALSE;
+	}
+
+	task_t * task = (task_t *)tasks->data;
+	if (task->wait_for != NULL) {
+		g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_BAD_VALUE, "Task is already waiting for %s.  Asked to wait for %s", task->wait_for, value);
+		return FALSE;
+	}
+
+	task->wait_for = g_strdup(value);
+	any_waitfors = TRUE;
 	return TRUE;
 }
 
@@ -582,12 +667,24 @@ normalize_name_length (void)
 	return;
 }
 
+static gboolean
+max_wait_hit (gpointer user_data)
+{
+	g_warning("Timing out at maximum wait of %d seconds.", max_wait);
+	g_main_loop_quit(global_mainloop);	
+	global_success = FALSE;
+	return FALSE;
+}
+
 static gchar * dbus_configfile = NULL;
+static gchar * dbus_daemon = NULL;
 
 static GOptionEntry general_options[] = {
-	{"dbus-config",  'd',   G_OPTION_FLAG_FILENAME,  G_OPTION_ARG_FILENAME,  &dbus_configfile, "Configuration file for newly created DBus server.  Defaults to '" DEFAULT_SESSION_CONF "'.", "config_file"},
-	{"bustle-data",  'b',   G_OPTION_FLAG_FILENAME,  G_OPTION_ARG_FILENAME,  &bustle_datafile, "A file to write out data from the bustle logger to.", "data_file"},
-	{"bustle-watch", 'w',   0,                       G_OPTION_ARG_CALLBACK,  bustle_watch,     "Defines a watch string for the bustle watcher task. (broken)", "filter"},
+	{"dbus-daemon",  0,     0,                       G_OPTION_ARG_FILENAME,  &dbus_daemon,     "Path to the DBus deamon to use.  Defaults to 'dbus-daemon'.", "executable"},
+	{"dbus-config",  'd',   0,                       G_OPTION_ARG_FILENAME,  &dbus_configfile, "Configuration file for newly created DBus server.  Defaults to '" DEFAULT_SESSION_CONF "'.", "config_file"},
+	{"bustle-monitor", 0,   0,                       G_OPTION_ARG_FILENAME,  &bustle_cmd,      "Path to the Bustle DBus Monitor to use.  Defaults to 'bustle-dbus-monitor'.", "executable"},
+	{"bustle-data",  'b',   0,                       G_OPTION_ARG_FILENAME,  &bustle_datafile, "A file to write out data from the bustle logger to.", "data_file"},
+	{"max-wait",     'm',   0,                       G_OPTION_ARG_INT,       &max_wait,        "The maximum amount of time the test runner will wait for the test to complete.  Default is 30 seconds.", "seconds"},
 	{NULL}
 };
 
@@ -597,6 +694,7 @@ static GOptionEntry task_options[] = {
 	{"ignore-return", 'r',  G_OPTION_FLAG_NO_ARG,     G_OPTION_ARG_CALLBACK,  option_noreturn, "Do not use the return value of the task to calculate whether the test passes or fails.", NULL},
 	{"invert-return", 'i',  G_OPTION_FLAG_NO_ARG,     G_OPTION_ARG_CALLBACK,  option_invert,   "Invert the return value of the task before calculating whether the test passes or fails.", NULL},
 	{"parameter",     'p',  0,                        G_OPTION_ARG_CALLBACK,  option_param,    "Add a parameter to the call of this utility.  May be called as many times as you'd like.", NULL},
+	{"wait-for",      'f',  0,                        G_OPTION_ARG_CALLBACK,  option_wait,     "A dbus-name that should appear on the bus before this task is started", "dbus-name"},
 	{NULL}
 };
 
@@ -606,6 +704,8 @@ main (int argc, char * argv[])
 	GError * error = NULL;
 	GOptionContext * context;
 
+	g_type_init();
+
 	context = g_option_context_new("- run multiple tasks under an independent DBus session bus");
 
 	g_option_context_add_main_entries(context, general_options, "dbus-runner");
@@ -626,21 +726,36 @@ main (int argc, char * argv[])
 		dbus_configfile = DEFAULT_SESSION_CONF;
 	}
 
+	if (dbus_daemon == NULL) {
+		dbus_daemon = "dbus-daemon";
+	}
+
+	if (bustle_cmd == NULL) {
+		bustle_cmd = "bustle-dbus-monitor";
+	}
+
 	gint dbus_stdout;
 	GPid dbus;
 	gchar * blank[1] = {NULL};
-	gchar * dbus_startup[] = {"dbus-daemon", "--config-file", dbus_configfile, "--print-address", NULL};
+	gchar * dbus_startup[] = {dbus_daemon, "--config-file", dbus_configfile, "--print-address", NULL};
 	g_spawn_async_with_pipes(g_get_current_dir(),
 	                         dbus_startup, /* argv */
 	                         blank, /* envp */
-	                         G_SPAWN_SEARCH_PATH, /* flags */
+	                         G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD, /* flags */
 	                         NULL, /* child setup func */
 	                         NULL, /* child setup data */
 							 &dbus, /* PID */
 	                         NULL, /* stdin */
 	                         &dbus_stdout, /* stdout */
 	                         NULL, /* stderr */
-	                         NULL); /* error */
+	                         &error); /* error */
+
+	if (error != NULL) {
+		g_critical("Unable to start dbus daemon: %s", error->message);
+		return 1;
+	}
+
+	guint dbus_watch = g_child_watch_add(dbus, dbus_watcher, NULL);
 
 	GIOChannel * dbus_io = g_io_channel_unix_new(dbus_stdout);
 	g_io_add_watch(dbus_io,
@@ -648,11 +763,16 @@ main (int argc, char * argv[])
 	               dbus_writes, /* func */
 	               NULL); /* func data */
 
+	if (max_wait > 0) {
+		g_timeout_add_seconds(max_wait, max_wait_hit, NULL);
+	}
+
 	global_mainloop = g_main_loop_new(NULL, FALSE);
 	g_main_loop_run(global_mainloop);
 
 	stop_bustling();
 
+	g_source_remove(dbus_watch); /* Let's not error when we want to kill it */
 	gchar * killline = g_strdup_printf("kill -9 %d", dbus);
 	g_spawn_command_line_sync(killline, NULL, NULL, NULL, NULL);
 	g_free(killline);
